package checkers.nonnull;

import java.io.IOException;
import java.util.Properties;

import javax.annotation.processing.ProcessingEnvironment;
import javax.annotation.processing.SupportedAnnotationTypes;
import javax.annotation.processing.SupportedSourceVersion;
import javax.lang.model.SourceVersion;
import javax.lang.model.element.AnnotationMirror;

import com.sun.source.tree.CompilationUnitTree;

import checkers.basetype.BaseTypeChecker;
import checkers.basetype.BaseTypeVisitor;
import checkers.source.SupportedLintOptions;
import checkers.types.AnnotatedTypeFactory;
import checkers.types.AnnotatedTypeMirror;
import checkers.types.AnnotationFactory;
import checkers.util.SimpleSubtypeRelation;

@SupportedAnnotationTypes({"checkers.quals.NonNull", "checkers.quals.Nullable"})
@SupportedLintOptions({"flow", "cast", "cast:redundant"})
@SupportedSourceVersion(SourceVersion.RELEASE_7)
public class NonNullChecker extends BaseTypeChecker {

    private AnnotationFactory annoFactory;
    private AnnotationMirror NULLABLE;
    private AnnotationMirror NONNULL;
    private SimpleSubtypeRelation relation;
    
    @Override
    public synchronized void init(ProcessingEnvironment processingEnv) {
        super.init(processingEnv);
        annoFactory = new AnnotationFactory(processingEnv);
        NULLABLE = this.annoFactory.fromName("checkers.quals.Nullable");
        NONNULL = this.annoFactory.fromName("checkers.quals.NonNull");
        relation = new SimpleSubtypeRelation(NONNULL, NULLABLE);
    }
    
    @Override
    public boolean isSubtype(AnnotatedTypeMirror lhs, AnnotatedTypeMirror rhs) {
        return this.relation.isSubtype(lhs, rhs);
    }
    
    @Override
    protected Properties getMessages(String fileName) throws IOException {
        return new Properties();
    }

    @Override
    protected BaseTypeVisitor<?, ?> getSourceVisitor(CompilationUnitTree root) {
        return new NonNullVisitor(this, root);
    }

    @Override
    protected String getDefaultSkipPattern() {
        return "";
    }

    @Override
    protected String getSuppressWarningsKey() {
        return "nonnull";
    }

    @Override
    public AnnotatedTypeFactory getFactory(ProcessingEnvironment env,
            CompilationUnitTree root) {
        return new NonNullAnnotatedTypeFactory(env, root);
    }
    
    
}
