package checkers.util;

import java.util.*;

import javax.lang.model.type.*;
import javax.lang.model.util.SimpleTypeVisitor6;

import checkers.quals.*;
import checkers.types.AnnotationLocation;

/**
 * A visitor for types that tracks locations as it descends into generic
 * types.
 */
public class LocationVisitor<R,P> 
    extends SimpleTypeVisitor6<R, P> {
    
    private List<Integer> loc = new ArrayList<Integer>();
    
    /**
     * @return the location of the currently visited type
     */
    public AnnotationLocation getCurrentLocation() {
        return AnnotationLocation.fromList(loc);
    }
    
    protected final R scan(TypeMirror t, P p) {
        return (t == null) ? null : visit(t, p);
    }
    
    protected R reduce(R r1,R r2) {
        return (r1 != null) ? r1 : r2;
    }
    
    protected R scanAndReduce(TypeMirror t, P p, R r) {
        return reduce(scan(t, p), r);
    }
    
    protected /*@Nullable*/ R scan(Iterable<? extends TypeMirror> types, P p) {
        boolean first = true;
        /*@Nullable*/ R r = null;
        for (TypeMirror t : types) {
            r = first ? scan(t, p) : scanAndReduce(t, p, r);
        }
        return r;
    }
    
    /**
     * This method needs to be called to visit 
     * type parameters.
     */
    @Override
    public R visitDeclared(DeclaredType t, P p) {
        R r = super.visitDeclared(t, p);
        final int depth = loc.size();
        loc.add(0);
        int i = 0;

        for (TypeMirror arg : t.getTypeArguments()) {
            loc.set(depth, i++);
            r = scanAndReduce(arg, p, r);
        }
        try {
            return r;
        } finally {
            // calls remove(int) not remove(Object)
            loc.remove(depth);
        }
    }
}
