package tests;

import org.junit.*;
import static org.junit.Assert.*;

import javax.tools.*;
import java.util.*;

import com.sun.source.tree.*;

/**
 * JUnit tests for the NonNull annotation checker.
 */
public class NonNullTest extends CheckerTest {

    public NonNullTest() {
        checkerName = "checkers.nonnull.NonNullChecker";
        checkerDir = "nonnull";
        System.setProperty("checkers.skipClasses", "");
    }

    /** Tests for expressions-sensitive analysis. */
    @Test public void testExpressions() {
        TestRun run = getTest("Expressions.java");
        assertSuccess(run);
        List<Diagnostic<? extends JavaFileObject>> list = run.getDiagnostics();
        assertDiagnostics(list, "Expressions.out");
    }

    /** Tests for flow-sensitive analysis. */
    @Test public void testFlow() {
        TestRun run = getTest("Flow.java");
        assertFailure(run);
        List<Diagnostic<? extends JavaFileObject>> list = run.getDiagnostics();
        assertDiagnostics(list, "Flow.out");
    }

    /** Tests for flow-sensitive analysis. */
    @Test public void testFlowSelf() {
        TestRun run = getTest("FlowSelf.java");
        assertSuccess(run);
    }

    /** Tests for flow-sensitive analysis. */
    @Test public void testFlowField() {
        TestRun run = getTest("FlowField.java");
        assertSuccess(run);
    }

    /** Tests for flow-sensitive analysis. */
    @Test public void testFlowAssignment() {
        TestRun run = getTest("FlowAssignment.java");
        assertSuccess(run);
    }

    /** Tests for flow-sensitive analysis. */
    @Test public void testMyException() {
        TestRun run = getTest("MyException.java");
        assertFailure(run);
        List<Diagnostic<? extends JavaFileObject>> list = run.getDiagnostics();
        assertDiagnostics(list, "MyException.out");
    }

    /** Tests that crashes do not occur for varargs methods. */
    @Test public void testVarargs() {
        TestRun run = getTest("Varargs.java");
        assertSuccess(run);
        List<Diagnostic<? extends JavaFileObject>> list = run.getDiagnostics();
        assertDiagnostics(list, "Varargs.out");
    }

    /** Tests the @Default annotation. */
    @Test public void testDefaultAnnotation() {
        TestRun run = getTest("DefaultAnnotation.java");
        assertFailure(run);
        assertDiagnostics(run.getDiagnostics(), "DefaultAnnotation.out");
    }

    @Test public void testArrayRefs() {
        TestRun run = getTest("ArrayRefs.java");
        assertFailure(run);
        assertDiagnostics(run.getDiagnostics(), "ArrayRefs.out");
    }

    @Test public void testArrayArgs() {
        TestRun run = getTest("ArrayArgs.java");
        assertFailure(run);
        assertDiagnostics(run.getDiagnostics(), "ArrayArgs.out");
    }

    @Test public void testDefaultFlow() {
        TestRun run = getTest("DefaultFlow.java");
        assertSuccess(run);
    }

    @Test public void testGenericArgs() {
        TestRun run = getTest("GenericArgs.java");
        assertSuccess(run);
    }

    /*
    @Test public void testMarino() {
        TestRun run = getTest("Marino.java");
        assertSuccess(run);
    }
    */
}
